from abc import ABC, abstractmethod
from math import floor
import os

import numpy as np

from model_server.image import GenericImageDataAccessor, InMemoryDataAccessor


class Model(ABC):

    def __init__(self, autoload=True, params=None):
        """
        Abstract base class for an inference model that uses image data as an input.

        :param autoload: automatically load model and dependencies into memory if True
        :param params: Dict[str, str] of arguments e.g. configuration files required to load model
        """
        self.autoload = autoload
        self.params = params
        self.loaded = False
        if not autoload:
            return None
        if self.load():
            self.loaded = True
        else:
            raise CouldNotLoadModelError()
        return None

    @abstractmethod
    def load(self):
        """
        Abstract method that carries out the expectedly time-consuming step of loading a model into memory
        :return: True if successful, else False
        """
        pass

    @abstractmethod
    def infer(self, img: GenericImageDataAccessor) -> (object, dict): # return json describing inference result
        pass

    def reload(self):
        self.load()


class ImageToImageModel(Model):
    """
    Abstract class for models that receive an image and return an image of the same size
    """

    @abstractmethod
    def infer(self, img: GenericImageDataAccessor) -> (GenericImageDataAccessor, dict):
        pass

class DummyImageToImageModel(ImageToImageModel):

    model_id = 'dummy_make_white_square'

    def load(self):
        return True

    def infer(self, img: GenericImageDataAccessor) -> (GenericImageDataAccessor, dict):
        super().infer(img)
        w = img.shape_dict['X']
        h = img.shape_dict['Y']
        result = np.zeros([h, w], dtype='uint8')
        result[floor(0.25 * h) : floor(0.75 * h), floor(0.25 * w) : floor(0.75 * w)] = 255
        return InMemoryDataAccessor(data=result), {'success': True}

class Error(Exception):
    pass

class CouldNotLoadModelError(Error):
    pass

class ParameterExpectedError(Error):
    pass